/*----------------------------------------------------------------------
  File    : invert.c
  Contents: invert a matrix and compute its determinant
  Author  : Christian Borgelt
  History : 2001.10.17 file created
            2003.06.11 computation of Moore-Penrose pseudo-inverse added
            2003.08.16 slight changes in error message output
            2007.02.13 adapted to modified module tabread
            2010.10.11 adapted to modified module tabread
            2010.12.11 adapted to a generic error reporting function
            2013.08.09 adapted to higher compiler warning level
            2013.08.13 adapted to definition of type DIMID
            2013.08.20 output format changed to significant digits
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include "escape.h"
#include "tabread.h"
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "invert"
#define DESCRIPTION "(pseudo)invert a matrix"
#define VERSION     "version 2.1 (2014.10.24)         " \
                    "(c) 2001-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in matrix.h */
#define E_STDIN     (-5)        /* double assignment of stdin */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_ALGO      (-9)        /* unknown algorithm */
#define E_SING     (-10)        /* singular matrix */
#define E_POSDEF   (-11)        /* matrix is not positive definite */
/* error codes  -16 to -18 defined in matrix.h */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_ALGO     -9 */  "unknown algorithm %d",
  /* E_SING    -10 */  "singular matrix",
  /* E_POSDEF  -11 */  "matrix is not positive definite",
  /*    -12 to -15 */  NULL, NULL, NULL, NULL,
  /* E_VALUE   -16 */  "%s:%d(%d): invalid value '%s'",
  /* E_FLDCNT  -17 */  "%s:%d(%d): wrong number of fields/columns",
  /* E_RECCNT  -18 */  "%s:%d(%d): wrong number of records/rows",
  /*           -19 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static TABREAD *tread  = NULL;  /* table/matrix reader */
static MATRIX  *mat    = NULL;  /* data matrix */
static MATRIX  *nonsqr = NULL;  /* non-square matrix */
static MATRIX  *transp = NULL;  /* transpose of non-square matrix */
static FILE    *out    = NULL;  /* output file */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (mat)   mat_delete(mat); \
  if (tread) trd_delete(tread, 1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of file to read */
  CCHAR   *fn_out  = NULL;      /* name of file to write */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  int     algo     = 'g';       /* algorithm to use */
  int     mode     = MAT_PARTPIV;  /* pivoting mode */
  int     cdet     = 0;         /* whether to compute determinant */
  int     digs     = 6;         /* number of significant digits */
  char    seps[4]  = " \n";     /* separators for output */
  double  det;                  /* determinant of matrix */
  DIMID   r, c;                 /* number of rows and columns */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-a#      algorithm to use                  "
                    "(default: %c)\n", algo);
    printf("         g: Gauss-Jordan elimination "
                    "(any regular matrix)\n");
    printf("         l: LU       decomposition   "
                    "(any regular matrix)\n");
    printf("         c: Cholesky decomposition   "
                    "(symmetric positive definite matrix)\n");
    printf("-d       compute determinant of the matrix "
                    "and its inverse\n");
    printf("-p       full pivoting (Gauss-Jordan only) "
                    "(default: partial pivoting)\n");
    printf("-o#      number of significant digits      "
                    "(default: %d)\n", digs);
    printf("-r#      record  separators                "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                "
                    "(default: \"#\")\n");
    printf("infile   matrix file to read (numbers only)\n");
    printf("outfile  file to write inverted matrix to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'a': algo   = (*s) ? *s++ : 'g';     break;
          case 'p': mode   = MAT_FULLPIV;           break;
          case 'd': cdet   = 1;                     break;
          case 'o': digs   = (int)strtol(s, &s, 0); break;
          case 'r': optarg = &recseps;              break;
          case 'f': optarg = &fldseps;              break;
          case 'b': optarg = &blanks;               break;
          case 'C': optarg = &comment;              break;
          default : error(E_OPTION, *--s);          break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (!strchr("glc012", algo)) error(E_ALGO, algo);
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read matrix file --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  k = mat_readx(&mat, tread, -1, -1);    /* read the matrix */
  if (k) error(k, TRD_INFO(tread));
  trd_delete(tread, 1);         /* close the input file and */
  tread = NULL;                 /* delete the table reader */
  r = mat_rowcnt(mat);          /* get the number of rows */
  c = mat_colcnt(mat);          /* and the number of columns */
  fprintf(stderr, "[%"DIMID_FMT" row(s),",    r);
  fprintf(stderr, " %"DIMID_FMT" column(s)]", c);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- invert the matrix --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "inverting matrix ... ");
  if (r != c) {                 /* if the matrix is not square */
    nonsqr = mat;               /* store it in another variable */
    transp = mat_create(c, r);  /* create a matrix for the transpose */
    mat    = mat_create(c, c);  /* and a matrix for the product */
    if (!transp || !mat) error(E_NOMEM);
    mat_transp(transp, nonsqr); /* transpose the input matrix */
    mat_mul(mat, transp, nonsqr);
    mat_delete(nonsqr);         /* multiply the input matrix */
    nonsqr = NULL;              /* with its transpose and */
  }                             /* delete the input matrix */
  switch (algo) {               /* evaluate the algorithm code */
    case 'g': case '0':         /* Gauss-Jordan elimination */
      if (mat_gjinv(mat, mat, mode, &det) != 0) error(E_SING);
      break;
    case 'l': case '1':         /* LU decomposition */
      if (mat_ludecom(mat, mat) != 0) error(E_SING);
      if (cdet) det = mat_ludet(mat);
      if (mat_luinv(mat, mat) != 0)   error(E_NOMEM);
      break;
    case 'c': case '2':         /* Cholesky decomposition */
      if (mat_chdecom(mat, mat) != 0) error(E_POSDEF);
      if (cdet) det = mat_chdet(mat);
      mat_chinv(mat, mat); mat_tr2sym(mat, mat, MAT_UPPER);
      break;
    default: error(E_ALGO, algo); break;
  }                             /* (solution is in matrix mat) */
  if (r != c) {                 /* if the matrix is not square */
    mat_mul(transp, mat, transp);
    mat_delete(mat);            /* compute the pseudo-inverse, */
    mat = transp; transp = NULL;/* store it in the result variable, */
  }                             /* and clear the transposed matrix */
  fprintf(stderr, "[%"DIMID_FMT" row(s),",    r);
  fprintf(stderr, " %"DIMID_FMT" column(s)]", c);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- write inverted matrix/determinant --- */
  t = clock();                  /* start the timer */
  if (fn_out && (strcmp(fn_out, "-") == 0)) fn_out = "";
  if (fn_out && *fn_out) { out = fopen(fn_out, "wb"); }
  else                   { out = stdout; fn_out = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_out);
  if (!out) error(E_FOPEN, fn_out);  /* open the output file */
  if (fldseps) seps[0] = (char)esc_decode(fldseps, NULL);
  if (recseps) seps[1] = (char)esc_decode(recseps, NULL);
  mat_write(mat, out, digs, seps);   /* write the matrix */
  if (((out == stdout) ? fflush(out) : fclose(out)) != 0)
    error(E_FWRITE, fn_out);    /* close the output file and */
  out = NULL;                   /* clear the file variable */
  fprintf(stderr, "[%"DIMID_FMT" row(s),",    c);
  fprintf(stderr, " %"DIMID_FMT" column(s)]", r);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  if (cdet) {                   /* if to print the determinant */
    printf("determinant of original matrix: ");
    printf("%-.*g", digs, (algo == 0) ? 1/det : det); putchar('\n');
    printf("determinant of inverted matrix: ");
    printf("%-.*g", digs, (algo == 0) ? det : 1/det); putchar('\n');
  }                             /* print both to avoid confusion */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
